from .models.k8s import Cluster, Node
from .models.region import Region
from .models.pagination import Pagination
from .models.functions import Container, Namespace, Function, Cron, Log
from .models.registry import Image
from .models.rdb import Privileges, User, Database, Instance
from .client_api import ClientAPI

from pydantic import validate_arguments
import boto3
import requests

import json
import os


class FunctionsClient(ClientAPI):

    def __init__(self, region: Region = Region.FrPar):
        super().__init__(name="functions", version="v1alpha2", region=region)

    # CONTAINERS

    @validate_arguments
    def list_containers(
        self,
        namespace_id: str = None,
        pagination: Pagination = Pagination(),
        ordering: Container.Ordering = Container.Ordering(),
        name: str = None,
        organization_id: str = None
    ): return [Container(**data) for data in json.loads(
            self.request("/containers", data={
                "name": name,
                "namespace_id": namespace_id,
                "organization_id": organization_id} | pagination.dict() | ordering.dict()))["containers"]]

    @validate_arguments
    def get_container(self, id: str):
        return Container(**json.loads(self.request(f"/containers/{id}")))

    @validate_arguments
    def create_container(self, container: Container):
        return Container(**json.loads(self.request("/containers", ClientAPI.Method.POST, data=container.dict())))

    @validate_arguments
    def deploy_container(self, id: str):
        return Container(**json.loads(self.request(f"/containers/{id}/deploy", ClientAPI.Method.POST)))

    @validate_arguments
    def update_container(self, id, container: Container):
        return Container(**json.loads(self.request(f"/containers/{id}", ClientAPI.Method.PATCH, data=container.dict())))

    @validate_arguments
    def delete_container(self, container: Container):
        return Container(**json.loads(self.request(
            f"/containers/{container.id}", ClientAPI.Method.DELETE, data=container.dict())))

    # NAMESPACES

    @validate_arguments
    def list_namespaces(
        self,
        pagination: Pagination = Pagination(),
        ordering: Namespace.Ordering = Namespace.Ordering(),
        name: str = None,
        organization_id: str = None
    ): return [Namespace(**data) for data in json.loads(
            self.request("/namespaces", data={
                "name": name, "organization_id": organization_id} | pagination.dict() | ordering.dict()))["namespaces"]]

    @validate_arguments
    def get_namespace(self, id: str):
        return Namespace(**json.loads(self.request(f"/namespaces/{id}")))

    @validate_arguments
    def create_namespace(self, namespace: Namespace):
        return Namespace(**json.loads(self.request("/namespaces", ClientAPI.Method.POST, data=namespace.dict())))

    @validate_arguments
    def update_namespace(self, namespace: Namespace):
        return Namespace(**json.loads(
            self.request(f"/namespaces/{namespace.id}", ClientAPI.Method.PATCH, data=namespace.dict())))

    @validate_arguments
    def delete_namespace(self, namespace: Namespace):
        return Namespace(**json.loads(
            self.request(f"/namespaces/{namespace.id}", ClientAPI.Method.DELETE, data=namespace.dict())))

    # FUNCTIONS

    @validate_arguments
    def list_functions(
        self,
        pagination: Pagination = Pagination(),
        ordering: Function.Ordering = Function.Ordering(),
        name: str = None,
        namespace_id: str = None,
        organization_id: str = None
    ): return [Function(**data) for data in json.loads(
            self.request("/functions", data={
                "name": name, "namespace_id": namespace_id, "organization_id": organization_id
                    } | pagination.dict() | ordering.dict()))["functions"]]

    @validate_arguments
    def get_function(self, id: str):
        return Function(**json.loads(self.request(f"/functions/{id}")))

    @validate_arguments
    def create_function(self, function: Function):
        return Function(**json.loads(self.request("/functions", ClientAPI.Method.POST, data=function.dict())))

    @validate_arguments
    def deploy_function(self, function: Function):
        return Function(**json.loads(self.request(
            f"/functions/{function.id}/deploy", ClientAPI.Method.POST, data=function.dict())))

    @validate_arguments
    def update_function(self, function: Function):
        return Function(**json.loads(
            self.request(f"/functions/{function.id}", ClientAPI.Method.PATCH, data=function.dict())))

    @validate_arguments
    def delete_function(self, function: Function):
        return Function(**json.loads(
            self.request(f"/functions/{function.id}", ClientAPI.Method.DELETE, data=function.dict())))

    # CRONS

    @validate_arguments
    def list_crons(
        self,
        pagination: Pagination = Pagination(),
        ordering: Cron.Ordering = Cron.Ordering(),
        application_id: str = None
    ): return [Cron(**data) for data in json.loads(
            self.request("/crons", data={
                "application_id": application_id} | pagination.dict() | ordering.dict()))["crons"]]

    @validate_arguments
    def get_cron(self, id: str):
        return Cron(**json.loads(self.request(f"/crons/{id}")))

    @validate_arguments
    def create_cron(self, cron: Cron):
        return Cron(**json.loads(self.request("/crons", ClientAPI.Method.POST, data=cron.dict())))

    @validate_arguments
    def update_cron(self, cron: Cron):
        return Cron(**json.loads(self.request(f"/crons/{cron.id}", ClientAPI.Method.PATCH, data=cron.dict())))

    @validate_arguments
    def delete_cron(self, cron: Cron):
        return Cron(**json.loads(self.request(f"/crons/{cron.id}", ClientAPI.Method.DELETE, data=cron.dict())))

    # LOGS

    @validate_arguments
    def list_logs(
        self,
        pagination: Pagination = Pagination(),
        ordering: Log.Ordering = Log.Ordering(),
        application_id: str = None
    ): return [Log(**data) for data in json.loads(
            self.request("/logs", data={
                "application_id": application_id} | pagination.dict() | ordering.dict()))["logs"]]


class RdbClient(ClientAPI):

    def __init__(self, region: Region = Region.FrPar):
        super().__init__(name="rdb", version="v1", region=region)

    # INSTANCES

    @validate_arguments
    def list_instances(
        self,
        name: str = None,
        project_id: str = None,
        organization_id: str = None,
        pagination: Pagination = Pagination(),
        ordering: Instance.Ordering = Instance.Ordering()
    ): return [Instance(**data) for data in json.loads(
            self.request("/instances", data={
                "organization_id": organization_id,
                "project_id": project_id,
                "name": name} | pagination.dict() | ordering.dict()))["instances"]]

    @validate_arguments
    def get_instance(
        self,
        instance_id
    ): return Instance(**json.loads(self.request(f"/instances/{instance_id}")))

    # DATABASES

    @validate_arguments
    def list_databases(
        self,
        instance: Instance,
        name: str = None,
        managed: bool = None,
        owner: str = None,
        ordering: Database.Ordering = Database.Ordering(),
        pagination: Pagination = Pagination()
    ): return [Database(**data) for data in json.loads(
            self.request(f"/instances/{instance.id}/databases", data={
                "owner": owner,
                "managed": managed,
                "name": name} | pagination.dict() | ordering.dict()))["databases"]]

    @validate_arguments
    def create_database(self, instance: Instance, database: Database):
        return Database(**json.loads(self.request(
            f"/instances/{instance.id}/databases", ClientAPI.Method.POST, data=database.dict())))

    @validate_arguments
    def delete_database(self, instance: Instance, database: Database):
        self.request(
            f"/instances/{instance.id}/databases/{database.name}", ClientAPI.Method.DELETE, data=database.dict())

    # USERS

    @validate_arguments
    def list_users(
        self,
        instance: Instance,
        name: str = None,
        ordering: User.Ordering = User.Ordering(),
        pagination: Pagination = Pagination()
    ): return [User(**data) for data in json.loads(
            self.request(f"/instances/{instance.id}/users", data={
                "name": name} | pagination.dict() | ordering.dict()))["users"]]

    @validate_arguments
    def create_user(
        self,
        instance: Instance,
        user: User,
        password: User.Password
    ):
        try:
            return User(**json.loads(self.request(
                f"/instances/{instance.id}/users", ClientAPI.Method.POST, data=user.dict() | {"password": password})))
        except requests.exceptions.HTTPError:
            raise ValueError("Unable to create user")

    @validate_arguments
    def update_user(
        self,
        instance: Instance,
        user: User,
        password: User.Password = None
    ): return User(**json.loads(self.request(
            f"/instances/{instance.id}/users/{user.name}", ClientAPI.Method.PATCH, data=user.dict() | {
                "password": password})))

    @validate_arguments
    def delete_user(self, instance: Instance, user: Database):
        self.request(
            f"/instances/{instance.id}/users/{user.name}", ClientAPI.Method.DELETE, data=user.dict())

    # PRIVILEGES

    @validate_arguments
    def list_privileges(
        self,
        instance: Instance,
        database_name: str = None,
        user_name: str = None,
        ordering: Privileges.Ordering = Privileges.Ordering(),
        pagination: Pagination = Pagination()
    ): return [Privileges(**data) for data in json.loads(
            self.request(f"/instances/{instance.id}/privileges", data={
                "database_name": database_name, "user_name": user_name
            } | pagination.dict() | ordering.dict()))["privileges"]]

    @validate_arguments
    def set_user_privileges(self, instance: Instance, privileges: Privileges):
        return Privileges(**json.loads(self.request(
            f"/instances/{instance.id}/privileges", method=ClientAPI.Method.PUT, data=privileges.dict())))


class RegistryClient(ClientAPI):

    def __init__(self, region: Region = Region.FrPar):
        super().__init__(name="registry", version="v1", region=region)

    # IMAGES

    @validate_arguments
    def list_images(
        self,
        name: str = None,
        pagination: Pagination = Pagination(),
        ordering: Image.Ordering = Image.Ordering(),
        application_id: str = None,
        organization_id: str = None,
        project_id: str = None
    ): return [Image(**data) for data in json.loads(
            self.request("/images", data={
                "application_id": application_id,
                "organization_id": organization_id,
                "project_id": project_id, "name": name} | pagination.dict() | ordering.dict()))["images"]]

    @validate_arguments
    def get_image(self, id: str):
        return Image(**json.loads(self.request(f"/image/{id}")))

    @validate_arguments
    def update_image(self, image: Image):
        return Image(**json.loads(self.request(f"/image/{image.id}", ClientAPI.Method.PATCH, data=image.dict())))

    @validate_arguments
    def delete_image(self, image: Image):
        return Image(**json.loads(self.request(f"/images/{image.id}", ClientAPI.Method.DELETE, data=image.dict())))


class K8sClient(ClientAPI):

    def __init__(self, region: Region = Region.FrPar):
        super().__init__(name="k8s", version="v1", region=region)

    # CLUSTERS

    @validate_arguments
    def list_clusters(
        self,
        name: str = None,
        status: Cluster.Status = Cluster.Status.Unknown,
        type: str = None,
        pagination: Pagination = Pagination(),
        ordering: Cluster.Ordering = Cluster.Ordering(),
        application_id: str = None,
        organization_id: str = None,
        project_id: str = None
    ): return [Cluster(**data) for data in json.loads(
            self.request("/clusters", data={
                "status": status.value,
                "type": type,
                "application_id": application_id,
                "organization_id": organization_id,
                "project_id": project_id,
                "name": name
            } | pagination.dict() | ordering.dict()))["clusters"]]

    @validate_arguments
    def download_kubeconfig(
        self,
        cluster_id: str
    ): return self.request(f"/clusters/{cluster_id}/kubeconfig?dl=1", data={
            "cluster_id": cluster_id
        }, to_json=False)

    # NODES

    @validate_arguments
    def list_nodes(
        self,
        cluster_id: str,
        pool_id: str = None,
        name: str = None,
        status: Node.Status = Node.Status.Unknown,
        pagination: Pagination = Pagination(),
        ordering: Node.Ordering = Node.Ordering(),
    ): return [Node(**data) for data in json.loads(
            self.request(f"/clusters/{cluster_id}/nodes", data={
                "status": status.value,
                "name": name,
                "pool_id": pool_id
            } | pagination.dict() | ordering.dict()))["nodes"]]

    @validate_arguments
    def get_node(
        self,
        node_id: str
    ): return Node(**json.loads(self.request(f"/nodes/{node_id}")))


class ObjectStorageClient:

    def __init__(self, region: Region = Region.FrPar):
        session = boto3.session.Session()
        endpoint_url = f'http://s3.{region.value}.scw.cloud'
        self.s3 = session.client(
            service_name='s3',
            region_name=region.value,
            use_ssl=True,
            endpoint_url=endpoint_url,
            aws_access_key_id=os.getenv("SCW_API_KEY_ID"),
            aws_secret_access_key=os.getenv("SCW_API_KEY")
        )

    def list_buckets(self):
        response = self.s3.list_buckets()
        return response['Buckets']

    def create_bucket(self, name: str):
        self.s3.create_bucket(Bucket=name)

    def enable_bucket_website(self, name: str, error_path: str = 'error.html', index_path: str = 'index.html'):
        self.s3.put_bucket_website(
            Bucket=name,
            WebsiteConfiguration={
                'ErrorDocument': {'Key': error_path},
                'IndexDocument': {'Suffix': index_path},
            }
        )

    def upload_file_to_bucket(self, file_path: str, bucket_name: str, file_name: str):
        self.s3.upload_file(file_path, bucket_name, file_name)



from pydantic import BaseModel
from typing import Dict, Any
from enum import Enum
from .region import Region


class Namespace(BaseModel):

    class Ordering(BaseModel):

        class OrderBy(Enum):

            CreatedAtAsc = "created_at_asc"
            CreatedAtDesc = "created_at_desc"
            NameAsc = "name_asc"
            NameDesc = "name_desc"

        class Config:
            use_enum_values = True

        order_by: OrderBy = OrderBy.CreatedAtAsc

    id: str = None
    name: str = None
    organization_id: str
    description: str = None
    environment_variables: Dict[str, Any] = None
    error_message: str = None
    region: Region = Region.FrPar
    registry_endpoint: str = None
    registry_namespace_id: str = None
    status: str = "ready"


class Log(BaseModel):

    class Ordering(BaseModel):

        class OrderBy(Enum):

            TimestampAsc = "timestamp_asc"
            TimestampDesc = "timestamp_desc"

        class Config:
            use_enum_values = True

        order_by: OrderBy = OrderBy.TimestampAsc

    id: str = None
    application_id: str = None
    timestamp: str = None
    message: str = None


class Function(BaseModel):

    class Privacy(Enum):

        Unknown = "unknown"
        Public = "public"
        Private = "private"

    class Runtime(Enum):

        Golang = "golang"
        Python = "python"
        Python3 = "python3"
        Node8 = "node8"
        Node10 = "node10"
        Node14 = "node14"

    class Ordering(BaseModel):

        class OrderBy(Enum):

            CreatedAtAsc = "created_at_asc"
            CreatedAtDesc = "created_at_desc"
            NameAsc = "name_asc"
            NameDesc = "name_desc"

        class Config:
            use_enum_values = True

        order_by: OrderBy = OrderBy.CreatedAtAsc

    class Config:
        use_enum_values = True

    id: str = None
    name: str = None
    status: str = None
    organization_id: str
    namespace_id: str
    description: str = None
    environment_variables: Dict[str, Any] = None
    error_message: str = None
    region: str = Region.FrPar
    registry_endpoint: str = None
    registry_namespace_id: str = None
    min_scale: int = 1
    max_scale: int = 20
    runtime: Runtime = Runtime.Golang
    handler: str = None
    memory_limit: int = None
    cpu_limit: int = None
    privacy: Privacy = Privacy.Unknown


class Cron(BaseModel):

    class Ordering(BaseModel):

        class OrderBy(Enum):

            CreatedAtAsc = "created_at_asc"
            CreatedAtDesc = "created_at_desc"

        class Config:
            use_enum_values = True

        order_by: OrderBy = OrderBy.CreatedAtAsc

    id: str = None
    application_id: str
    schedule: str
    args: Dict[str, Any] = None


class Container(BaseModel):

    class Privacy(Enum):

        Unknown = "unknown"
        Public = "public"
        Private = "private"

    class Ordering(BaseModel):

        class OrderBy(Enum):

            CreatedAtAsc = "created_at_asc"
            CreatedAtDesc = "created_at_desc"
            NameAsc = "name_asc"
            NameDesc = "name_desc"

        class Config:
            use_enum_values = True

        order_by: OrderBy = OrderBy.CreatedAtAsc

    class Config:
        use_enum_values = True

    id: str = None
    namespace_id: str
    name: str
    timeout: int
    environment_variables: dict = None
    min_scale: int = None
    max_scale: int = None
    memory_limit: int = None
    cpu_limit: int = None
    privacy: Privacy = Privacy.Unknown
    description: str = None
    registry_image: str = None
    max_concurrency: int = None
    domain_name: str = None



from datetime import datetime
from typing import List
from pydantic import BaseModel
from enum import Enum


class Image(BaseModel):

    class Privacy(Enum):

        Unknown = "unknown"
        Public = "public"
        Private = "private"

    class Ordering(BaseModel):

        class OrderBy(Enum):

            CreatedAtAsc = "created_at_asc"
            CreatedAtDesc = "created_at_desc"
            NameAsc = "name_asc"
            NameDesc = "name_desc"

        class Config:
            use_enum_values = True

        order_by: OrderBy = OrderBy.CreatedAtAsc

    class Config:
        use_enum_values = True

    id: str
    namespace_id: str
    name: str
    status: str
    status_message: str = None
    visibility: str
    size: int
    created_at: datetime
    updated_at: datetime
    tags: List[str]



from .models.region import Region

from typing import Any, Dict

import requests
import json
import os
from enum import Enum


class ClientAPI:

    class Method(Enum):
        GET = "GET"
        POST = "POST"
        PATCH = "PATCH"
        DELETE = "DELETE"
        PUT = "PUT"

    class Error(Exception):

        def __init__(self, *args: object):
            super().__init__(*args)

    def __init__(self, name: str, version: str, region: Region = Region.FrPar):
        token = os.getenv("SCW_API_KEY")
        if not token:
            raise Exception("SCW_API_KEY must be defined as environment variable")
        self.API_url = f"https://api.scaleway.com/{name}/{version}/regions/{region.value}"
        self.headers = {
            "accept": "application/json",
            "X-Auth-Token": token,
            "Content-Type": "application/json"
        }

    def request(self, url: str, method: Method = Method.GET, data: Dict[str, Any] = {}, to_json=True):
        if method == ClientAPI.Method.GET:
            res = requests.get(self.API_url + url, headers=self.headers, params=data)
        elif method == ClientAPI.Method.POST:
            res = requests.post(self.API_url + url, headers=self.headers, data=json.dumps(ClientAPI._clean_dict(data)))
        elif method == ClientAPI.Method.PATCH:
            res = requests.patch(self.API_url + url, headers=self.headers, data=json.dumps(ClientAPI._clean_dict(data)))
        elif method == ClientAPI.Method.DELETE:
            res = requests.delete(self.API_url + url, headers=self.headers)
        elif method == ClientAPI.Method.PUT:
            res = requests.put(self.API_url + url, headers=self.headers, data=json.dumps(ClientAPI._clean_dict(data)))
        else:
            raise Exception(f"Unhandled method: {method}")
        if not res.ok:
            raise ClientAPI.Error(res.json()["message"])
        if method != ClientAPI.Method.DELETE and to_json:
            return json.dumps(res.json(), indent=4, sort_keys=True)
        else:
            return res.content

    @classmethod
    def _clean_dict(cls, d):
        if not isinstance(d, dict):
            return d
        return dict((k, cls._clean_dict(v)) for k, v in d.items() if v is not None)



from pydantic import BaseModel, validator
from enum import Enum
from typing import List
from datetime import datetime
import re

from pydantic.types import PositiveInt


class Privileges(BaseModel):

    class Permission(Enum):

        ReadOnly = "readonly"
        ReadWrite = "readwrite"
        All = "all"
        Custom = "custom"
        _None = "none"

    class Ordering(BaseModel):

        class OrderBy(Enum):

            UserNameAsc = "user_name_asc"
            UserNameDesc = "user_name_desc"
            DatabaseNameAsc = "database_name_asc"
            DatabaseNameDesc = "database_name_desc"

        class Config:
            use_enum_values = True

        order_by: OrderBy = OrderBy.UserNameAsc

    class Config:
        use_enum_values = True

    permission: Permission = Permission.ReadOnly
    database_name: str
    user_name: str


class User(BaseModel):

    class Ordering(BaseModel):

        class OrderBy(Enum):

            NameAsc = "name_asc"
            NameDesc = "name_desc"
            IsAdminAsc = "is_admin_asc"
            IsAdminDec = "is_admin_desc"

        class Config:
            use_enum_values = True

        order_by: OrderBy = OrderBy.NameAsc

    class Config:
        use_enum_values = True

    class Password(str):

        def __new__(cls, v):
            if not re.match(r"^(?=\S*[a-z])(?=\S*[A-Z])(?=\S*\d)(?=\S*[^\w\s])\S{8,}$", v):
                raise ValueError(
                    "Password must have at least one upper character, one lower character, "
                    "one special character, one number and length must be 8 or more")
            return str.__new__(cls, v)

    @validator("name")
    def check_name(cls, v):
        if not re.match(r"^[\w\$\-]{1,63}$", v) or not v[0].isalpha() or v.startswith("_rdb"):
            raise ValueError(
                f"User name ({v}) length must be between 1 and 63 characters, first character must be an alphabet "
                "character (a-zA-Z), and cannot start with '_rdb', only a-zA-Z0-9_$- characters are accepted")
        return v

    name: str
    is_admin: bool = False


class Instance(BaseModel):

    class Endpoint(BaseModel):
        ip: str
        port: PositiveInt
        name: str = None

    class Ordering(BaseModel):

        class OrderBy(Enum):

            CreatedAtAsc = "created_at_asc"
            CreatedAtDesc = "created_at_desc"
            NameAsc = "name_asc"
            NameDesc = "name_desc"
            Region = "region"
            StatusAsc = "status_asc"
            StatusDec = "status_desc"

        class Config:
            use_enum_values = True

        order_by: OrderBy = OrderBy.CreatedAtAsc

    class Config:
        use_enum_values = True

    id: str
    name: str = None
    status: str
    status_message: str = None
    created_at: datetime
    tags: List[str]
    endpoint: Endpoint = None


class Database(BaseModel):

    class Ordering(BaseModel):

        class OrderBy(Enum):

            NameAsc = "name_asc"
            NameDesc = "name_desc"
            SizeAsc = "size_asc"
            SizeDesc = "size_desc"

        class Config:
            use_enum_values = True

        order_by: OrderBy = OrderBy.NameAsc

    class Config:
        use_enum_values = True

    name: str = None
    owner: str = None
    status: str = None
    size: int = None
